import csv
import os
import urllib.request
import shutil
import zipfile

from cleo import Command
from cleo.helpers import option

from rgov.commands import paths


class UpdateIndexCommand(Command):

    name = "reindex"
    description = "Update the campground id search index"

    help = """ The <info>reindex</info> command downloads facility data from
recreation.gov and builds a searchable database from that.

By default, the database is built with campground names and ids. The<comment>
--with-descriptions </comment>option includes campground descriptions in the
index, which can be useful for searching for attributes like park name, state,
or city. The size of the database with just names is ~128kb, and is ~9mb with
the descriptions included.

"""

    options = [
        option(
            "with-descriptions",
            "w",
            "Whether to include descriptions in the search database.",
        )
    ]

    def ensure_download_dir_exists(self):
        os.makedirs(paths.download_folder, exist_ok=True)

    def request_ridb_data(self):
        urllib.request.urlretrieve(paths.download_url, paths.download_path)

    def unzip_ridb_data(self):
        with zipfile.ZipFile(paths.download_path, "r") as zip_ref:
            zip_ref.extractall(paths.download_folder)

    def generate_index(self, descriptions):
        """
        Reads from the recreation.gov facilities csv file
        and writes a new csv that includes only the attributes
        that are relevant to the program. This includes
        campground names and ids and optionally, descriptions,
        which provide for better search capabilities but take
        more disk space (around 9mb vs ~130kb). Some data cleaning
        is performed, such as removing html tags and fixing
        capitalization.

        The relevant Recreation.gov facilities csv columns are:
        Column 0 = Facility ID
        Column 5 = Facility Name
        Column 6 = Facility Description
        Column 7 = Facility Type
        Column 19 = Reservable (true or false)

        The output csv is organized as:
        Column 0 = Facility ID
        Column 1 = Facility Name
        Column 2 = Facility Description (if --with-descriptions)
        """
        os.makedirs(paths.data_folder, exist_ok=True)

        if descriptions is True:
            import re

            def cleanhtml(raw_html):
                cleanr = re.compile("<.*?>")
                cleantext = re.sub(cleanr, "", raw_html)
                return cleantext

        with open(paths.facilities_csv_path, "r") as i:
            reader = csv.reader(i)
            with open(paths.index_path, "w") as o:
                writer = csv.writer(o)
                for row in reader:
                    # filter for reservable campgrounds with a non-empty name entry
                    if row[7] == "Campground" and row[19] == "true" and row[5]:
                        if descriptions is True:
                            writer.writerow(
                                [row[0], row[5].lower(), cleanhtml(row[6]).lower()]
                            )
                        else:
                            writer.writerow([row[0], row[5].lower()])

    def delete_temp_files(self):
        """
        Removes the project-local downlaod folder
        and its contents.
        """
        shutil.rmtree(paths.download_folder)

    def handle(self):
        descriptions = self.option("with-descriptions")

        steps = [
            ("Ensuring download directory exists", self.ensure_download_dir_exists),
            (f"Downloading data from {paths.download_url}", self.request_ridb_data),
            ("Unzipping", self.unzip_ridb_data),
            ("Generating index", self.generate_index),
            ("Deleting temporary files", self.delete_temp_files),
        ]

        if not self.confirm(
            "Are you sure you want to regenerate the campsite id index?", False
        ):
            return

        for step in steps:
            format = f"<question>{step[0]}...</question>"
            self.line("")
            self.write(format)
            if step[0] == "Generating index":
                step[1](descriptions)  # needs argument
            else:
                step[1]()
            self.overwrite(format + f"<comment>done</comment>")

        self.line("")



import daemon
import datetime
import time
from subprocess import run, Popen

from cleo import Command
from cleo.helpers import option, argument

from rgov.commands import check_command

# TODO add logging for errors and to show that the daemon is running
# TODO add to or better the notification method(s) (e.g. signal, email, twitter, phone, etc)
# TODO show in notifications all campgrounds with availability, not just one or something vague

class DaemonCommand(check_command.CheckCommand):

    name = "daemon"
    description = "Start a daemon that checks for availablity automatically"

    help = """The <info>daemon</info> command automatically checks campground(s)
availability on the dates of stay every 5 minutes as a background process. Like
the check command, the daemon command defaults to the current date with a three
nights stay. The following example checks for a three night stay starting on
October 12, 2021 at Laguna:

<info>pandcamp daemon 232279 -date 10-12-2021 -length 3</info>

Campground ids can be found using the <info>search</info> command.

The <comment>command</comment> option runs a given shell command in the event
one or more of the campgrounds checked are found to have available sites. This
can be used, for example, to send an email with msmtp.

The <comment>--notifier</comment> option specifies a notification program to be
used (e.g. herbe or notify-send) to send a system notification. The notifier
assumes that the format of the notification program is <comment><notification
program> "notification text"</comment>.

Be default, the daemon exits if availability is found, to avoid excess
notifications. This can be disabled with the <comment>--forever</comment>
option.

"""

    arguments = [
        argument("id", "The campground id(s) to check", multiple=True),
    ]

    options = [
        option(
            "length",
            "l",
            "Length of stay in nights [3 nights]",
            flag=False,
            value_required=True,
        ),
        option("forever", "f", "Continue to run even after finding open sites"),
        option(
            "date",
            "d",
            "The date of your arrival, in mm-dd-yyyy format [today]",
            flag=False,
            value_required=True,
        ),
        option(
            "notifier",
            "N",
            "The notifier to run if an availability is found [notify-send]",
            flag=False,
            value_required=True,
        ),
        option(
            "command",
            "c",
            "The command to run if an availability is found",
            flag=False,
            value_required=True),
    ]

    def handle(self):
        campground_ids = self.argument("id")

        if self.option("notifier"):
            notifier = self.option("notifier")
        else:
            notifier = "notify-send"

        self.line("<fg=magenta>Pandcamp daemon started.</fg=magenta>")

        with daemon.DaemonContext():
            while True:
                for campground_id in campground_ids:
                    try:
                        campground_name, available_sites = self.main(campground_id)
                    except UnboundLocalError:
                        run([notifier, "Pandcamp", f"{campground_id} - invalid id!"])
                        continue
                    if len(available_sites) >= 1:
                        if self.option("notifier"):
                            run([notifier, f"{campground_name} - sites found!"])
                        if self.option("command"):
                            command = self.option("command")
                            Popen(command, shell=True)
                        if self.option("forever"):
                            pass

                time.sleep(300)



from os.path import join
from pathlib import Path


def get_project_root() -> Path:
    return Path(__file__).parent.parent.parent


project_root = get_project_root()

# downloads
download_url = "https://ridb.recreation.gov/downloads/RIDBFullExport_V1_CSV.zip"
download_folder = join(project_root, "assets")
download_path = join(download_folder, "RIDBExport.zip")
facilities_csv_path = join(download_folder, "Facilities_API_v1.csv")

# indexes
data_folder = join(project_root, "data")
index_path = join(data_folder, "campgrounds_index.csv")



from cleo.helpers import argument, option

from rgov.commands import check_command

class CheckCommand(check_command.CheckCommand):

    name = "check"
    description = "Check a campground for availability"

    arguments = [
        argument("id", "The campground id(s) to check", multiple=True)
    ]
    options = [
        option(
            "length",
            "l",
            "Length of stay in nights [3 nights]",
            flag=False,
            value_required=True,
        ),
        option("url", "u", "Display the campground url(s)"),
        option(
            "date",
            "d",
            "The date of your arrival, in mm-dd-yyy format [today]",
            flag=False,
            value_required=True,
        ),
    ]

    help = """The <info>check</info> command checks campground(s) for availability on the
dates of stay and prints the results to the terminal. The following example
checks for a three night stay starting on October 12, 2021 at Laguna:

<info>pandcamp check 232279 -date 10-12-2021 -length 3</info>

Campground ids can be found using <info>search</info> command. To check multiple
campgrounds, separate the campground ids with spaces:

<info>pandcamp check 232279 232278</info>

Unless otherwise specified, the command checks for the current date and a
length of stay of three days. The <comment>--url</comment> option prints the
url along with the results for quickly navigating to the reservation web page.

"""

    def handle(self) -> int:
        campground_ids = self.argument("id")

        for campground_id in campground_ids:
            try:
                campground_name, available_sites = self.main(campground_id)
            except UnboundLocalError:
                invalid_id_line = (f"<fg=red> \u2022 <info>{campground_id}</info> "
                                   f"is not a valid campground id</fg=red>")
                self.line(invalid_id_line)
                continue

            num_sites = len(available_sites)

            # if id is valid
            if 1 <= num_sites <= 12:
                sitelist_sorted = sorted(available_sites)
                result_string = " ".join(sitelist_sorted)
            elif num_sites > 12:
                result_string = f"{num_sites} sites available"
            else:
                result_string = "<fg=yellow>No sites available</fg=yellow>"

            result_line = (f"<question> \u2022 {campground_name}</question> - "
                           f"<fg=cyan>{result_string}</fg=cyan>")
            url_line = (f"<fg=green>   "
                        "https://www.recreation.gov/camping/campgrounds/"
                        f"{campground_id}/availability</fg=green>")

            self.line(result_line)
            if self.option("url"):
                self.line(url_line)



from cleo import Application

from rgov.commands import check, search, reindex, daemon

commands = [
    check.CheckCommand(),
    search.SearchCommand(),
    reindex.UpdateIndexCommand(),
    daemon.DaemonCommand(),
]

application = Application()

[application.add(command) for command in commands]


def main() -> int:
    return application.run()


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rgov', 'rgov.commands']

package_data = \
{'': ['*']}

install_requires = \
['cleo>=0.8.1,<0.9.0',
 'fake-useragent>=0.1.11,<0.2.0',
 'python-daemon>=2.3.0,<3.0.0',
 'requests>=2.25.1,<3.0.0']

entry_points = \
{'console_scripts': ['rgov = rgov.application:main']}

setup_kwargs = {
    'name': 'rgov',
    'version': '0.1.1',
    'description': '',
    'long_description': None,
    'author': 'Jordan Sweet',
    'author_email': 'jsbmgcontact@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import csv

from cleo import Command
from cleo.helpers import option, argument

from rgov.commands import paths


class SearchCommand(Command):

    name = "search"
    description = "Search for campground ids"
    arguments = [argument("terms", "The terms to search for", multiple=True)]
    options = [
        option("descriptions", "d", "Search descriptions"),
    ]

    help = """The <info>search</info> command searches text to find campground
ids. Multiple search terms can be separated by spaces, to search in an
orderless fashion, or in quotes, to search for exact matches.

By default, the keywords are matched against campground names. The<comment>
--descriptions </comment>option searches against campground descriptions, but
this options requires that the search index be rebuilt (see the
<info>reindex</info> command). This can be done with:

<info>pandcamp reindex --with-descriptions</info>.
"""

    def handle(self):
        """Search campground names and descriptions and print their name and id."""
        search_terms = self.argument("terms")
        search_terms_lowercase = [term.lower() for term in search_terms]
        num_search_terms = len(search_terms)

        # Read the index csv file line by line. If a match is found, the
        # corresponding name and id is printed. If multiple search words, all
        # must match the search string, in any order. The target_column
        # variable specifies which column to search in. Column 1 is names, and
        # column 2 contains the descriptions.
        if self.option("descriptions"):
            target_column = 2  # the search target column
        else:
            target_column = 1
        with open(paths.index_path, "r") as i:
            reader = csv.reader(i)
            try:
                for row in reader:
                    count = 0
                    for term in search_terms_lowercase:
                        if term in row[target_column]:
                            count += 1
                            if count == num_search_terms:
                                self.line(f"<question>{row[1].title()}</question>"
                                          f" - <info>{row[0]}</info>")

            # This error occurs when description option is passed but the csv
            # isn't built with the descriptions included.
            except IndexError:
                error_string = ("<fg=red>Description searches "
                                "are disabled.</fg=red>\nEnable with "
                                "<info>pandcamp reindex --with-descriptions</info>.")
                self.line(error_string)



import csv
import datetime
import json
import requests

from cleo import Command
from fake_useragent import UserAgent

from rgov.commands import paths


class CheckCommand(Command):
    base_url = "https://www.recreation.gov/api/camps/availability/campground/"
    headers = {"User-Agent": UserAgent().random}
    human_time_format = "%m/%d/%y"
    request_time_format = "%Y-%m-%dT00:00:00.000Z"
    response_time_format = "%Y-%m-%dT00:00:00Z"

    def main(self, campground_id):
        """Requests campground data and returns the name and a list of
        available sites for the given dates.

        Raises:
            UnboundLocalError: When the campground id is not recognized
        """
        # Get the name of the campground from the id. Iterates through the csv
        # until the id matches and assign the campground_name variable to the
        # corresponding name
        with open(paths.index_path, "r") as f:
            reader = csv.reader(f)
            for row in reader:
                if campground_id in row[0]:
                    campground_name = row[1].title()

        # Gather date and length of stay input, otherwise use the default
        # values
        if self.option("date"):
            date_input = self.option("date")
            try:
                arrival_date = datetime.datetime.strptime(date_input, "%m-%d-%Y")
            except ValueError:
                date_error_line = (f'"{date_input}" is not a valid date'
                              ' in the form mm-dd-yyyy')
                raise ValueError(date_error_line)
        else:
            arrival_date = datetime.date.today()
        if self.option("length"):
            length_input = self.option("length")
            try:
                length_of_stay = int(length_input)
            except ValueError:
                length_error_line = (f'"{length_input}" is not an integer')
                raise ValueError(length_error_line)
        else:
            length_of_stay = 3

        # Creates a list of request dates stored in the variable
        # "request_dates". The site requires that request dates be on the first
        # of the month, and in the format specified in the variable
        # request_time_format, so if the dates span into a new calendar month,
        # we need to generate more requests with the proper dates.
        request_dates = []
        first_day = arrival_date.replace(day=1)
        last_day = arrival_date + datetime.timedelta(days=length_of_stay)
        for month in range(first_day.month, (last_day.month + 1)):
            rqst_date = first_day.replace(month=month)
            rqst_date_fmtd = rqst_date.strftime(self.request_time_format)
            request_dates.append(rqst_date_fmtd)

        # Generate a list of the days thay we intend to camp on. This starts on
        # the arrival_date and spans the number of days in length_of_stay. This
        # is formatted how the response dates are formatted, so that we can
        # select values from the responses data using this list.
        stay_dates = []
        stay_range = range(length_of_stay)
        for i in stay_range:
            date = arrival_date + datetime.timedelta(days=i)
            date_formatted = date.strftime(self.response_time_format)
            stay_dates.append(date_formatted)

        # Requests campground availability date for the request dates created
        # above, and stores each response in a list. If the response is empty
        # it may be an invalid id, so the function will return with
        # "INVALID_ID".
        url = f"{self.base_url}{campground_id}/month?"
        responses = []
        for date in request_dates:
            data = requests.get(url, {"start_date": date}, headers=self.headers)
            data_loaded = json.loads(data.text)
            try:
                campsite_data = data_loaded["campsites"]
            except KeyError as e:
                raise UnboundLocalError("Invalid ID")
            campsite_values = campsite_data.values()
            responses.append(campsite_values)

        # Create a list of all sites for the campground. Iterate through the
        # responses, appending sites to the unavailable_sites list if they are
        # unavailable on any of the stay dates. Create a list of all sites that
        # aren't unavailable, and return that.
        all_sites = [site["site"] for site in responses[0]]
        unavailable_sites = []
        for response in responses:
            for site in response:
                for date in stay_dates:
                    if date in site["availabilities"]:
                        if site["availabilities"][date] not in "Available":
                            unavailable_sites.append(site["site"])
                            break
        available_sites = [site for site in all_sites if site not in unavailable_sites]

        # raise ValueError if campground_name is empty
        try:
            return campground_name, available_sites
        except UnboundLocalError as e:
           raise UnboundLocalError('Invalid campground ID')
